import pandas as pd
import os, uuid, json, re, shutil

# Import a custom function to create the date heirarchies
import PBI_dashboard_creator.create_date_hrcy as PBI_date_hr


def add_csv(dashboard_path, data_path):

  # generate a random id for the data set
	dataset_id = str(uuid.uuid4())


	# extract bits of names for later
	path_end = os.path.basename(data_path)
	split_end = os.path.splitext(path_end)

	dataset_name = split_end[0]
	dataset_extension =split_end[1]


	report_name = os.path.basename(dashboard_path)

  # Reverse slash directions bc windows is stooooooopid
	data_path_reversed = data_path.replace('/', '\\')
	


	# file paths
	semantic_model_folder = os.path.join(dashboard_path, f'{report_name}.SemanticModel' )
	definitions_folder = os.path.join(semantic_model_folder, "definition")
	model_path = os.path.join(definitions_folder, 'model.tmdl')
	temp_model_path = os.path.join(dashboard_path, 'model2.tmdl')

	relationships_path = os.path.join(definitions_folder, "relationships.tmdl")
	diagram_layout_path = os.path.join(semantic_model_folder, 'diagramLayout.json')

	tables_folder = os.path.join(definitions_folder, 'tables')
	dataset_file_path = os.path.join(tables_folder, f'{dataset_name}.tmdl')

	# create a tables folder if it doesn't already exist
	if not os.path.exists(tables_folder):
		os.makedirs(tables_folder)


	# load dataset using pandas
	dataset = pd.read_csv(data_path)





	# add dataset to diagramLayout file ---------------------------------------------------------------------
	with open(diagram_layout_path,'r') as file:
		diagram_layout = json.load(file)


	# add all this junk to describe the table's "nodes"
	diagram_layout["diagrams"][0]["nodes"].append( 
        {
          "location": {
            "x": 0,
            "y": 0
          },
          "nodeIndex": dataset_name,
          "nodeLineageTag": dataset_id,
          "size": {
            "height": 300,
            "width": 234
          },
          "zIndex": 0
        }
      )

   
	# write to file
	with open(diagram_layout_path,'w') as file:
		json.dump(diagram_layout, file, indent = 2)





	# modify model.tdml file -------------------------------------------------------------------------------------
	with open(temp_model_path, 'w') as tmp:
		with open(model_path, "r") as file:
			for line in file.readlines():

				# check to see if the line is the one we want
				m = re.search("(?<=annotation PBI_QueryOrder = ).*", line)

				# if it is, read the list of datasets and append a new one in

				if m is not None:
					# execute the tmdl code to make a python list

					# execute the code (including local and global scopes)
					# source: https://stackoverflow.com/questions/41100196/exec-not-working-inside-function-python3-x
					exec(f'query_order_list = {m.group(0)}', locals(), globals())

					# add the dataset using python method then write back  to line
					query_order_list.append(dataset_name)
					line = f'annotation PBI_QueryOrder = {query_order_list}\n'

					# write back the line to a temporary file

				tmp.write(line)

			# append the dataset name at the end of the file
			tmp.write(f"\n\nref table {dataset_name}")


  # Replace the model file with the temp file we created
	shutil.move(temp_model_path, model_path)


	# Data model file --------------------------------------------------------------------------
    
    # sink inital header stuff about dataset
	with open(dataset_file_path, 'w') as file:
		file.write(f'table {dataset_name}\n\tlineageTag: {dataset_id}\n\n')

    # read in the dataset
    # compare how pandas manages to do this in a single line 
    # and Power BI requires 40 lines of code and modifying multiple files to do the same thing
    # in case you needed evidence of how dummmmmmmbbb Power BI, Power querry and M are.....



	for col in dataset:

		# Loop through the dataset and find dates
		for value in dataset[col][0:100]:
			m = re.search("^\d{4}-\d{2}-\d{2}$", str(value))

			if m is not None:
				#print(f"{col}: This column is probably a date!")

				# change the data type in the panda dataframe
				dataset[col] = pd.to_datetime(dataset[col], format = "%Y-%m-%d")

				# create a date heirarchy table
				file_id = PBI_date_hr.create_date_hr(col_name = col,
				 dataset_name = dataset_name,
				  report_name = report_name, 
				  dashboard_path = dashboard_path )
				break


	col_names = []
	col_deets = []


  # loop through columns and write specs out to model file
	for col in dataset:

		# loop through the values in a column to see if it contains dates
		# Loop through the dataset and find dates
		for value in dataset[col][0:100]:
			m = re.search("^\d{4}-\d{2}-\d{2}$", str(value))

			if m is not None:
				#print(f"{col}: This column is probably a date!")

				# change the data type in the panda dataframe
				dataset[col] = pd.to_datetime(dataset[col], format = "%Y-%m-%d")

				# create a date heirarchy table
				file_id = create_date_hr(col_name = col, 
					dataset_name = dataset_name, 
					report_name = report_name,
					 dashboard_path = dashboard_path )
				break

		# add the column's name to a set for later
		col_names.append(col)

		# record more details in a different set

		col_id = str(uuid.uuid4())


    # For numbers, we're not distinguishing between integers (int64)
    # and numbers (double)
		if dataset[col].dtype == "int64" or dataset[col].dtype == "float64":

			# record more details in a different set
			col_deets.append(f'{{"{col}", type number}}')


			with open(dataset_file_path, 'a') as file:
				file.write(f'\tcolumn {col}\n')
				file.write('\t\tdataType: double\n')
				#file.write('\t\tformatString: 0\n')
				file.write(f'\t\tlineageTag: {col_id}\n')
				file.write('\t\tsummarizeBy: sum\n')
				file.write(f'\t\tsourceColumn: {col}\n\n')
				file.write('\t\tannotation SummarizationSetBy = Automatic\n\n')
				file.write('\t\tannotation PBI_FormatHint = {"isGeneralNumber":true}\n\n')


    # strings ------------------------------------------------
		if dataset[col].dtype == "object":

			# record more details in a different set
			col_deets.append(f'{{"{col}", type text}}')

			with open(dataset_file_path, 'a') as file:
				file.write(f'\tcolumn {col}\n')
				file.write('\t\tdataType: string\n')
				file.write(f'\t\tlineageTag: {col_id}\n')
				file.write('\t\tsummarizeBy: none\n')
				file.write(f'\t\tsourceColumn: {col}\n\n')
				file.write('\t\tannotation SummarizationSetBy = Automatic\n\n')

 
 		# dates ----------------------------------------------
		if dataset[col].dtype == "datetime64[ns]":

			# create a relationship id
			relationship_id = str(uuid.uuid4())

			# record more details in a different set
			col_deets.append(f'{{"{col}", type date}}')

			with open(dataset_file_path, 'a') as file:
				file.write(f'\tcolumn {col}\n')
				file.write(f'\t\tdataType: dateTime\n')
				file.write(f'\t\tformatString: Long Date\n')
				file.write(f'\t\tlineageTag: {col_id}\n')
				file.write('\t\tsummarizeBy: none\n')
				file.write(f'\t\tsourceColumn: {col}\n\n')
				file.write(f'\t\tvariation Variation\n')
				file.write('\t\t\tisDefault\n')
				file.write(f'\t\t\trelationship: {relationship_id}\n')
				file.write(f"\t\t\tdefaultHierarchy: LocalDateTable_{file_id}.'Date Hierarchy'\n\n")
				file.write('\t\tannotation SummarizationSetBy = Automatic\n\n')
				file.write('\t\tannotation UnderlyingDateTimeDataType = Date\n\n')


			# create a new file to define the relationship
			with open(relationships_path, "a") as file:
				file.write(f'relationship {relationship_id}\n')
				file.write(f'\tjoinOnDateBehavior: datePartOnly\n')
				file.write(f'\tfromColumn: {dataset_name}.{col}\n')
				file.write(f'\ttoColumn: LocalDateTable_{file_id}.Date\n\n')

			# Append the date table to the model.tmdl file
			with open(model_path, "a") as file:
				file.write(f'\nref table LocalDateTable_{file_id}')






	# write out M code 
	# bc we're stilllllllll not done.....
	with open(dataset_file_path, 'a') as file:
		file.write(f'\tpartition {dataset_name} = m\n')
		file.write('\t\tmode: import\n\t\tsource =\n\t\t\t\tlet\n')
		file.write(f'\t\t\t\t\tSource = Csv.Document(File.Contents("{data_path_reversed}"),[Delimiter=",", Columns={len(dataset.columns)}, Encoding=1252, QuoteStyle=QuoteStyle.None]),\n')
		file.write('\t\t\t\t\t#"Promoted Headers" = Table.PromoteHeaders(Source, [PromoteAllScalars=true]),\n')
		file.write(f'\t\t\t\t\t#"Replaced Value" = Table.ReplaceValue(#"Promoted Headers","NA",null,Replacer.ReplaceValue,{{\"{'", "'.join(col_names)}\"}}),\n')
		file.write(f'\t\t\t\t\t#"Changed Type" = Table.TransformColumnTypes(#"Replaced Value",{{{', '.join(map(str, col_deets))}}})\n')
		file.write('\t\t\t\tin\n\t\t\t\t\t#"Changed Type"\n\n')
		file.write('\tannotation PBI_ResultType = Table\n\n\tannotation PBI_NavigationStepName = Navigation\n\n')










# Create a new blank .pbir dashboard
import os, uuid, shutil, json

from pathlib import Path
from importlib import resources



def create_new_dashboard(parent_dir, report_name):


	#create a new logical id field
	# see this for explanation of what a UUID is: https://stackoverflow.com/a/534847 
	report_logical_id = str(uuid.uuid4())
	sm_logical_id = str(uuid.uuid4())

	# define page name
	page1_name = "page1"

    
    # Define file paths ------------------------------------------------------------------------------------
	# Outer level directory --------------------------------------------------------------------------------
	project_folder_path = os.path.join(parent_dir, report_name)

	pbip_file_path = os.path.join(project_folder_path, f'{report_name}.pbip')


	## Report folder -----------------------------------------------------------------
	report_folder_path = os.path.join(project_folder_path, f'{report_name}.Report')
	platform_file_path = os.path.join(report_folder_path,  ".platform")
	pbir_file_path = os.path.join(report_folder_path, 'definition.pbir')


	### definition folder -------------------------------------------------------------------------------------
	report_definition_folder = os.path.join(report_folder_path, 'definition')

	pages_folder = os.path.join(report_definition_folder, 'pages')
	pages_file_path = os.path.join(pages_folder, "pages.json")


	page1_folder = os.path.join(pages_folder, page1_name)
	page1_json_path = os.path.join(page1_folder, "page.json")



	## report_name.SemanticModel folder ----------------------------------------------------------------------------
	semantic_model_folder_path = os.path.join(project_folder_path, f'{report_name}.SemanticModel')
	sm_platform_file_path = os.path.join(semantic_model_folder_path, ".platform")




	# check to make sure parent directory exists
	if not os.path.exists(parent_dir):
		raise ValueError("The parent directory doesn't exist! Please create it and try again!")


	# make sure a report folder doesn't already exist
	if os.path.exists(project_folder_path):
		raise ValueError("Sorry a report with that name already exists! Please use a different report name or parent directory and try again")
	
	


	# Transfer all the blank dashboard files from the package resources ---------------------------------------------------

	traversable = resources.files("PBI_dashboard_creator.dashboard_resources")
	
	with resources.as_file(traversable) as path:
		shutil.copytree(path, project_folder_path)


	# Change file names -----------------------------------------------------------------------------------------------------
	os.rename(os.path.join(project_folder_path, "blank_template.Report"), report_folder_path)
	os.rename(os.path.join(project_folder_path, "blank_template.SemanticModel"), os.path.join(project_folder_path, f'{report_name}.SemanticModel'))

	os.rename(os.path.join(project_folder_path, "blank_template.pbip"), pbip_file_path)

	os.rename(os.path.join(project_folder_path, f'{report_name}.Report/definition/pages/915e09e5204515bccac2'), os.path.join(project_folder_path, f'{report_name}.Report/definition/pages/{page1_name}'))



	# Modify files --------------------------------------------------------------------



	## top level -----------------------------------------------------------------------


	# .pbip file
	with open(pbip_file_path,'r') as file:
		pbip_file = json.load(file)
    
    # modify the report path
	pbip_file["artifacts"][0]["report"]["path"] = f'{report_name}.Report'

	# write to file
	with open(pbip_file_path,'w') as file:
		json.dump(pbip_file, file, indent = 2)





	## report folder -----------------------------------------------------------------

	# .platform file
	with open(platform_file_path,'r') as file:
		platform_file = json.load(file)

    
    # modify the display name
	platform_file["metadata"]["displayName"] = f'{report_name}'


	# update the unique UUID
	platform_file["config"]["logicalId"] = report_logical_id

	# write to file
	with open(platform_file_path,'w') as file:
		json.dump(platform_file, file, indent = 2)



	#.pbir file
	with open(pbir_file_path,'r') as file:
		pbir_file = json.load(file)

    # modify the display name
	pbir_file["datasetReference"]["byPath"]["path"] = f'../{report_name}.SemanticModel'


	# write to file
	with open(pbir_file_path,'w') as file:
		json.dump(pbir_file, file, indent = 2)



	### definition folder --------------------------------------------------------

	# pages.json
	with open(pages_file_path,'r') as file:
		pages_file = json.load(file)


	pages_file["pageOrder"] = [page1_name]
	pages_file["activePageName"] = page1_name

	# write to file
	with open(pages_file_path,'w') as file:
		json.dump(pages_file, file, indent = 2)


	#### page 1 folder
	with open(page1_json_path,'r') as file:
		page1_json = json.load(file)


	page1_json["name"] = page1_name

	# write to file
	with open(page1_json_path,'w') as file:
		json.dump(page1_json, file, indent = 2)



	## Semantic model folder ----------------------------------------------------------------
	# .platform file
	with open(platform_file_path,'r') as file:
		platform_file = json.load(file)

    
    # modify the display name
	platform_file["metadata"]["displayName"] = f'{report_name}'


	# update the unique UUID
	platform_file["config"]["logicalId"] = sm_logical_id

	# write to file
	with open(platform_file_path,'w') as file:
		json.dump(platform_file, file, indent = 2)



	### definition folder






















import os, json


# , chart_type

def add_chart(dashboard_path,
              page_id, 
              chart_id, 
              chart_type,
              data_source, 
              chart_title,
              x_axis_title,
              y_axis_title,
              x_axis_var, 
              y_axis_var, 
              y_axis_var_aggregation_type, 
              x_position, 
              y_position, 
              height, 
              width ):


	# file paths -------------------------------
	report_name = os.path.basename(dashboard_path)

	pages_folder = os.path.join(dashboard_path, f'{report_name}.Report/definition/pages')
	page_folder_path = os.path.join(pages_folder, page_id)

	visuals_folder = os.path.join(page_folder_path, "visuals")
	new_visual_folder = os.path.join(visuals_folder, chart_id)
	visual_json_path = os.path.join(new_visual_folder, "visual.json")







	# checks ---------------------------------------------------------

	# page exists? 
	if os.path.isdir(page_folder_path) is not True:
		raise NameError(f"Couldn't find the page folder at {page_folder_path}")

	# chart id unique? 
	if os.path.isdir(new_visual_folder) is True:
		raise ValueError(f'A visual with that chart_id already exists! Try using a different chart_id')

	else: 
		os.makedirs(new_visual_folder)





	#glob.glob(dashboard_path)



	# define the json for the new chart
	chart_json = {
  "$schema": "https://developer.microsoft.com/json-schemas/fabric/item/report/definition/visualContainer/1.3.0/schema.json",
  "name": chart_id,
  "position": {
    "x": x_position,
    "y": y_position,
    "z": 0,
    "height": height,
    "width": width,
    "tabOrder": 0
  },
  "visual": {
    "visualType": chart_type,
    "query": {
      "queryState": {
        "Category": {
          "projections": [
            {
              "field": {
                "Column": {
                  "Expression": {
                    "SourceRef": {
                      "Entity": data_source
                    }
                  },
                  "Property": x_axis_var
                }
              },
              "queryRef": f"{data_source}.{x_axis_var}",
              "nativeQueryRef": x_axis_var,
              "active": True
            }
          ]
        },
        "Y": {
          "projections": [
            {
              "field": {
                "Aggregation": {
                  "Expression": {
                    "Column": {
                      "Expression": {
                        "SourceRef": {
                          "Entity": data_source
                        }
                      },
                      "Property": y_axis_var
                    }
                  },
                  "Function": 0
                }
              },
              "queryRef": f"{y_axis_var_aggregation_type}({data_source}.{y_axis_var})",
              "nativeQueryRef": f"{y_axis_var_aggregation_type} of {y_axis_var}"
            }
          ]
        }
      },
      "sortDefinition": {
        "sort": [
          {
            "field": {
              "Aggregation": {
                "Expression": {
                  "Column": {
                    "Expression": {
                      "SourceRef": {
                        "Entity": data_source
                      }
                    },
                    "Property": y_axis_var
                  }
                },
                "Function": 0
              }
            },
            "direction": "Descending"
          }
        ],
        "isDefaultSort": True
      }
    },
    "objects": {
      "categoryAxis": [
        {
          "properties": {
            "titleText": {
              "expr": {
                "Literal": {
                  "Value": f"'{x_axis_title}'"
                }
              }
            }
          }
        }
      ],
      "valueAxis": [
        {
          "properties": {
            "titleText": {
              "expr": {
                "Literal": {
                  "Value": f"'{y_axis_title}'"
                }
              }
            }
          }
        }
      ]
    },
    "visualContainerObjects": {
      "title": [
        {
          "properties": {
            "text": {
              "expr": {
                "Literal": {
                  "Value": f"'{chart_title}'"
                }
              }
            }
          }
        }
      ]
    },
    "drillFilterOtherVisuals": True
  }
}

	# Write out the new json 
	with open(visual_json_path, "w") as file:
		json.dump(chart_json, file, indent = 2)


